unit Octree;

interface

uses
  Windows, Messages, dglOpenGL, Globals;

Const
  MAX_TRIANGELS_IN_NODE = 1000;

type
  PNode = ^TNode;
  TNode = object
    pos       : TVertex;
    size      : glfloat;
    numv      : Integer;
    numc      : Byte;
    poly      : array of PPolygon;
    children  : array of PNode;
    smallest  : Boolean;
    function HowManyPolygons(const polygons : array of PPolygon;Add:Boolean):Integer;
    function PolygonIn(p:PPolygon):Boolean;
    Procedure Divide(const polygons : array of PPolygon);
    procedure DrawPolygons;
    procedure Check;
    procedure Draw;
    procedure Clear;
  end;
  TOctree = class
  public
    mainnode  : PNode;
    procedure drawOctree(drawPol,drawNodes:Boolean);
    Constructor Create(const polygons : array of PPolygon);
    Destructor Destroy;
  end;

var
  numsmallnodes:Integer = 0;     //Um den Loding-Status anzuzeigen

implementation

uses Unit1;          //Um den Loding-Status anzuzeigen
                     //Ansonsten nicht ntig
Constructor TOctree.Create(const polygons : array of PPolygon);
var
  i,j:Integer;
  maxSize,minSize:Glfloat;
begin
  inherited Create;
  maxsize:=polygons[0].v[0].ver.x;
  minsize:=polygons[0].v[0].ver.x;
  for i:=0 to High(polygons) do
  for j:=0 to 2 do
  with polygons[i].v[j].ver do
  begin
 	  if x > maxSize then maxSize:=x;
    if y > maxSize then maxSize:=y;
    if z > maxSize then maxSize:=z;

    if x < minSize then minSize:=x;
    if y < minSize then minSize:=y;
    if z < minSize then minSize:=z;
  end;
  numsmallnodes:=0;
  New(mainnode);
  MainNode.pos.x:=(maxSize+minSize)/2;
  MainNode.pos.y:=(maxSize+minSize)/2;
  MainNode.pos.z:=(maxSize+minSize)/2;
  MainNode.size:=(maxSize-minSize)/2;
  MainNode.smallest:=FALSE;
  MainNode.Divide(polygons);
end;

function TNode.PolygonIn(p:PPolygon):Boolean;
var
  i:Integer;
begin
  Result := FALSE;
  for i := 0 to 2 do
  if not Result then
  with p.v[i].ver do
  begin
    if(x >= pos.x - size) and
      (y >= pos.y - size) and
      (z >= pos.z - size) and
      (x <= pos.x + size) and
      (y <= pos.y + size) and
      (z <= pos.z + size) then Result:=TRUE;
  end;
end;

function TNode.HowManyPolygons(const polygons : array of PPolygon; Add:Boolean):Integer;
var
  i:Integer;
begin
  numv:=0;
  for i:=0 to High(polygons) do
  if PolygonIn(polygons[i])then
  begin
    Inc(numv);
    if Add then
    begin
      setLength(poly,numv);
      poly[numv-1]:=polygons[i];
    end;
  end;
  Result:=numv;
end;

Procedure TNode.Divide(const polygons : array of PPolygon);
var
  TempNodes:array[0..7] of PNode;
  i:Integer;
begin
  for i:=0 to 7 do New(TempNodes[i]);
  numc:=0;
  TempNodes[0].pos.x:=pos.x-size/2;
  TempNodes[0].pos.y:=pos.y+size/2;
  TempNodes[0].pos.z:=pos.z-size/2;

  TempNodes[1].pos.x:=pos.x+size/2;
  TempNodes[1].pos.y:=pos.y+size/2;
  TempNodes[1].pos.z:=pos.z-size/2;

  TempNodes[2].pos.x:=pos.x+size/2;
  TempNodes[2].pos.y:=pos.y+size/2;
  TempNodes[2].pos.z:=pos.z+size/2;

  TempNodes[3].pos.x:=pos.x-size/2;
  TempNodes[3].pos.y:=pos.y+size/2;
  TempNodes[3].pos.z:=pos.z+size/2;

  TempNodes[4].pos.x:=pos.x-size/2;
  TempNodes[4].pos.y:=pos.y-size/2;
  TempNodes[4].pos.z:=pos.z-size/2;

  TempNodes[5].pos.x:=pos.x+size/2;
  TempNodes[5].pos.y:=pos.y-size/2;
  TempNodes[5].pos.z:=pos.z-size/2;

  TempNodes[6].pos.x:=pos.x+size/2;
  TempNodes[6].pos.y:=pos.y-size/2;
  TempNodes[6].pos.z:=pos.z+size/2;

  TempNodes[7].pos.x:=pos.x-size/2;
  TempNodes[7].pos.y:=pos.y-size/2;
  TempNodes[7].pos.z:=pos.z+size/2;
  for i:=0 to 7 do
  begin
    TempNodes[i].size:=size/2;
    TempNodes[i].smallest:=FALSE;
  end;
  for i:=0 to 7 do
  begin
    if TempNodes[i].HowManyPolygons(polygons,FALSE)>0 then
    begin
      inc(numc);
      setLength(children,numc);
      children[numc-1]:=TempNodes[i];
      if children[numc-1].numv<MAX_TRIANGELS_IN_NODE then
      begin
        inc(numsmallnodes);
        children[numc-1].smallest:=TRUE;
        children[numc-1].HowManyPolygons(polygons,TRUE);
        Form1.ProgressBar1.Position:=
          25+(75*numsmallnodes div((W*W)div(MAX_TRIANGELS_IN_NODE div 8)));
      end;
    end else Dispose(TempNodes[i]);
  end;
  for i:=0 to numc-1 do
    if children[i].smallest=FALSE then children[i].Divide(polygons);
end;

procedure TNode.draw;
var
  i:Integer;
begin
  if Frustum.IsBoxWithin(
     pos.x,pos.y,pos.z,
     size,size,size)=FALSE then exit;
  with pos do
  begin
    glBegin(GL_LINES);
      glVertex3f(x-size,y-size,z-size);
      glVertex3f(x+size,y-size,z-size);
      glVertex3f(x-size,y+size,z-size);
      glVertex3f(x+size,y+size,z-size);
      glVertex3f(x-size,y-size,z+size);
      glVertex3f(x+size,y-size,z+size);
      glVertex3f(x-size,y+size,z+size);
      glVertex3f(x+size,y+size,z+size);

      glVertex3f(x+size,y+size,z+size);
      glVertex3f(x+size,y-size,z+size);
      glVertex3f(x+size,y+size,z-size);
      glVertex3f(x+size,y-size,z-size);
      glVertex3f(x-size,y+size,z+size);
      glVertex3f(x-size,y-size,z+size);
      glVertex3f(x-size,y+size,z-size);
      glVertex3f(x-size,y-size,z-size);

      glVertex3f(x+size,y+size,z+size);
      glVertex3f(x+size,y+size,z-size);
      glVertex3f(x+size,y-size,z+size);
      glVertex3f(x+size,y-size,z-size);
      glVertex3f(x-size,y-size,z+size);
      glVertex3f(x-size,y-size,z-size);
      glVertex3f(x-size,y+size,z+size);
      glVertex3f(x-size,y+size,z-size);
    glEnd;
  end;
  if Length(children) > 0 then
  for i:=0 to Length(children)-1 do children[i].draw;
end;

procedure TNode.drawPolygons;
var
  i,j:Integer;
begin
  if numv = 0 then exit;
  glBegin(GL_TRIANGLES);
    for i := 0 to numv-1 do
    for j := 0 to 2 do
    with poly[i].v[j]^ do
    begin
      glNormal3fv(@n);
      glTexCoord2f(u,v);glVertex3fv(@ver);
    end;
  glEnd;
end;

procedure TNode.check;
var
  i:Integer;
begin
  if Frustum.IsBoxWithin(
     pos.x,pos.y,pos.z,
     size,size,size)=TRUE then
  begin
    if smallest then drawPolygons;
    for i:=0 to numc-1 do
      if not smallest then children[i].check;
  end;
end;

procedure TOctree.drawOctree(drawPol,drawNodes:Boolean);
begin
  if drawPol then
  begin
    glEnable(GL_TEXTURE_2D);
    glEnable(GL_LIGHTING);
    MainNode.check;
  end;
  if drawNodes then
  begin
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    glColor3f(1,1,0);
    MainNode.draw;
    glColor3f(1,1,1);
  end;
end;


procedure TNode.clear;
var
  i:Integer;
begin
  if Length(children)>0 then
  for i:=0 to Length(children)-1 do children[i].clear;
  setLength(children,0);
  Dispose(@self);
end;

Destructor TOctree.Destroy;
begin
  MainNode.clear;
  inherited Destroy;
end;

end.
