unit Unit1;

interface

uses
	Windows, Classes, Forms, Graphics, ComCtrls, StdCtrls, SysUtils, ExtCtrls, Controls,
  FrustumC, dglOpenGL, Octree, Globals;

type
  TForm1 = class(TForm)
    Timer1: TTimer;
    ProgressBar1: TProgressBar;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure FormKeyDown(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure Timer1Timer(Sender: TObject);
    procedure FormKeyUp(Sender: TObject; var Key: Word;
      Shift: TShiftState);
    procedure FormMouseMove(Sender: TObject; Shift: TShiftState; X,
      Y: Integer);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure FormResize(Sender: TObject);
	public
		DC:HDC;
		RC:HGLRC;
		procedure OnIdle(Sender:TObject;var Done:Boolean);
		procedure LoadTerrain(filename:String);
    procedure DoInit;
	private
	end;

Const
	W=256;      //Bitmap Size
  W2=W DIV 2;

var
  Form1: TForm1;
  ter:array[-W2..W2-1,-W2..W2-1]of TFullVertex;
  polygons:array of PPolygon;
  drawNodes:Boolean=TRUE;
  drawPol:Boolean=TRUE;
  useOctree:Boolean=TRUE;
  keys:array [0..255] of Boolean;
  octree:TOctree;

  terraintex:GLuInt;

  Frame,TimeFactorTmp,
  mx,my,tx,ty:Integer;
	RX,RY,TimeFactor:Glfloat;
	pos:TVertex;
  Ready:Boolean;

implementation

{$R *.dfm}

procedure pressedKeys;
begin
 	If keys[VK_UP] then
 	begin
 		pos.x:=Pos.x-sin(RY*PI/180)*cos(RX*PI/180)*TimeFactor;
   	pos.y:=pos.y+sin(RX*PI/180)*TimeFactor;
 		pos.z:=Pos.z+cos(RY*PI/180)*cos(RX*PI/180)*TimeFactor;
 	end;
	If keys[VK_DOWN] then
 	begin
 		pos.x:=Pos.x+sin(RY*PI/180)*cos(RX*PI/180)*TimeFactor;
   	pos.y:=pos.y-sin(RX*PI/180)*TimeFactor;
 		pos.z:=Pos.z-cos(RY*PI/180)*cos(RX*PI/180)*TimeFactor;
 	end;
	If keys[VK_LEFT] then
 	begin
   	pos.x:=Pos.x-sin((-RY-90)*PI/180)*TimeFactor;
   	pos.z:=Pos.z-cos((-RY-90)*PI/180)*TimeFactor;
 	end;
 	If keys[VK_RIGHT] then
 	begin
   	pos.x:=Pos.x+sin((-RY-90)*PI/180)*TimeFactor;
   	pos.z:=Pos.z+cos((-RY-90)*PI/180)*TimeFactor;
 	end;
end;

function GetNormal(v1,v2,v3:TVertex):TVertex;
var
  v4,v5, nor: TVertex;
  lenght:GLfloat;
begin
  v4.x:=V1.x-V2.x;
  v4.y:=V1.y-V2.y;
  v4.z:=V1.z-V2.z;
  v5.x:=V2.x-V3.x;
  v5.y:=V2.y-V3.y;
  v5.z:=V2.z-V3.z;
  nor.x:=v4.y*v5.z-v4.z*v5.y;
  nor.y:=v4.z*v5.x-v4.x*v5.z;
  nor.z:=v4.x*v5.y-v4.y*v5.x;
  lenght:=sqrt(nor.x*nor.x+nor.y*nor.y+nor.z*nor.z);
  nor.x:=nor.x/lenght;
  nor.y:=nor.y/lenght;
  nor.z:=nor.z/lenght;
  Result:=nor;
end;

procedure TForm1.LoadTerrain(filename:String);
var
  b:TBitmap;
  i,j,k:Integer;
  Vnormals:array[0..7] of Tvertex;
  lenght:Glfloat;
begin
  b:=TBitmap.Create;
  b.LoadFromFile(Filename);
  for i:=0 to W-1 do
  for j:=0 to W-1 do
  begin
    ter[i-W2,j-W2].ver.x:=(i-W2)/15*(256/W);
    ter[i-W2,j-W2].ver.y:=getRValue(b.Canvas.Pixels[i,W-1-j])/70;
    ter[i-W2,j-W2].ver.z:=(j-W2)/15*(256/W);
    ter[i-W2,j-W2].u:=i/(W-1);
    ter[i-W2,j-W2].v:=j/(W-1);
  end;
  b.Free;

  for i:=-W2 to W2-2 do
  for j:=-W2 to W2-2 do
  begin
    if (i>-W2)and(i<W2-1)and (j>-W2)and(j<W2-1) then
    begin
      Vnormals[0]:=GetNormal(ter[i,j].ver,ter[i+1,j  ].ver,ter[i+1,j-1].ver);
      Vnormals[1]:=GetNormal(ter[i,j].ver,ter[i+1,j-1].ver,ter[i  ,j-1].ver);
      Vnormals[2]:=GetNormal(ter[i,j].ver,ter[i+1,j+1].ver,ter[i+1,j  ].ver);
      Vnormals[3]:=GetNormal(ter[i,j].ver,ter[i  ,j+1].ver,ter[i+1,j+1].ver);
      Vnormals[4]:=GetNormal(ter[i,j].ver,ter[i-1,j+1].ver,ter[i  ,j+1].ver);
      Vnormals[5]:=GetNormal(ter[i,j].ver,ter[i-1,j  ].ver,ter[i-1,j+1].ver);
      Vnormals[6]:=GetNormal(ter[i,j].ver,ter[i-1,j-1].ver,ter[i-1,j  ].ver);
      Vnormals[7]:=GetNormal(ter[i,j].ver,ter[i  ,j-1].ver,ter[i-1,j-1].ver);
    end;
    ter[i,j].n:=Vnormals[0];
    for k:=1 to 7 do
    begin
      ter[i,j].n.x:=(ter[i,j].n.x+Vnormals[k].x);
      ter[i,j].n.y:=(ter[i,j].n.y+Vnormals[k].y);
      ter[i,j].n.z:=(ter[i,j].n.z+Vnormals[k].z);
    end;
    lenght:=sqrt(sqr(ter[i,j].n.x)+sqr(ter[i,j].n.y)+sqr(ter[i,j].n.z));
    ter[i,j].n.x:=ter[i,j].n.x/lenght;
    ter[i,j].n.y:=ter[i,j].n.y/lenght;
    ter[i,j].n.z:=ter[i,j].n.z/lenght;

    setLength(polygons,Length(polygons)+1);
    polygons[High(polygons)].v[0]:=@ter[i  ,j  ];
    polygons[High(polygons)].v[1]:=@ter[i+1,j+1];
    polygons[High(polygons)].v[2]:=@ter[i+1,j  ];

    setLength(polygons,Length(polygons)+1);
    polygons[High(polygons)].v[0]:=@ter[i  ,j  ];
    polygons[High(polygons)].v[1]:=@ter[i  ,j+1];
    polygons[High(polygons)].v[2]:=@ter[i+1,j+1];
  end;
  for i:=-W2 to W2-1 do
  for j:=-W2 to W2-1 do
  begin
    if (i=-W2)and(j>-W2)and(j<W2-1) then ter[i,j].n:=ter[i+1,j].n;
    if (i= W2-1)and(j>-W2)and(j<W2-1) then ter[i,j].n:=ter[i-1,j].n;
    if (j=-W2)and(i>-W2)and(i<W2-1) then ter[i,j].n:=ter[i,j+1].n;
    if (j= W2-1)and(i>-W2)and(i<W2-1) then ter[i,j].n:=ter[i,j-1].n;
  end;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  application.OnIdle:=onidle;
end;

procedure TForm1.DoInit;
begin
  ProgressBar1.Position:=0;
  Label1.Caption:='Initalizing OpenGL ...';
  label1.Refresh;
  InitOpenGL;
  ProgressBar1.Position:=5;
  Label1.Caption:='Initalizing Devices ...';
  label1.Refresh;
  DC:=GetDC(Form1.Handle);
  RC:=CreateRenderingContext(DC,[opDoubleBuffered],32,0,0,0,0,0);
	ActivateRenderingContext(DC,RC);
  glClearColor(0,0,0,0);
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_CULL_FACE);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);
  ProgressBar1.Position:=12;
  Label1.Caption:='Loading Terrain ...';
  label1.Refresh;
  LoadTerrain('Terrain.bmp');
  glPolygonMode(GL_FRONT_AND_BACK,GL_FILL);
  ProgressBar1.Position:=25;
  Label1.Caption:='Calculating Octree ...';
  label1.Refresh;
  octree:=TOctree.Create(polygons);
  pos.x:=0;
  pos.y:=-2;
  pos.z:=-1;
  rx:=-10;
  ProgressBar1.Visible:=FALSE;
  Label1.Visible:=FALSE;
  onResize(self);
  Ready:=TRUE;
end;

procedure drawFullLandScape;
var
  i,j:Integer;
begin
  glEnable(GL_LIGHTING);
  glBegin(GL_TRIANGLES);
    for i:=0 to High(polygons) do
    for j:=0 to 2 do
    with polygons[i].v[j]^ do
    begin
      glNormal3fv(@n);
      glTexCoord2f(u,v);glVertex3fv(@ver);
    end;
  glEnd;
end;

procedure Tform1.onidle(Sender:TObject;var Done:Boolean);
begin
  If not Ready then DoInit;
  Done:=FALSE;
  TimeFactorTmp:=GetTickCount;
  inc(Frame);
	glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
  pressedKeys;
  glLoadIdentity;
 	glRotatef(RX,1,0,0);
 	glRotatef(RY,0,1,0);
 	glTranslatef(pos.x,pos.y,pos.z);
  if useOctree then
  begin
 	  Frustum.Calculate;
    octree.drawOctree(drawPol,drawNodes);
  end else drawFullLandScape;
	SwapBuffers(DC);
	TimeFactor:=(GetTickCount-TimeFactorTmp)/100;
end;

procedure TForm1.FormResize(Sender: TObject);
begin
  if (ExtensionsRead and ImplementationRead)=FALSE then Exit;
  glViewPort(0,0,clientwidth,clientheight);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity;
	gluPerspective(45,clientwidth/clientheight,0.01,100);
  glMatrixMode(GL_MODELVIEW);
  ActivateRenderingContext(DC,RC);
end;

procedure TForm1.FormKeyDown(Sender: TObject; var Key: Word;
	Shift: TShiftState);
begin
  keys[key]:=TRUE;
  If keys[ord('P')] then drawPol:=not drawPol;
  If keys[ord('N')] then drawNodes:=not drawNodes;
  If keys[VK_SPACE] then useOctree:=not useOctree;
  If keys[VK_ESCAPE] then Close;
end;

procedure TForm1.Timer1Timer(Sender: TObject);
var
  s:String;
begin
  s:='Octree FPS: '+inttostr(frame);
  if useOctree then s:=s+' working with ' else s:=s+' without ';
  s:=s+' Octree';
  form1.Caption:=s;
  frame:=0;
end;

procedure TForm1.FormKeyUp(Sender: TObject; var Key: Word;
  Shift: TShiftState);
begin
	keys[key]:=FALSE;
end;

procedure TForm1.FormMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
  If ssRight in Shift then
  begin
  	RY:=RY+((x-mx)/10);
    RX:=RX+((y-my)/10);
    If RX>90 then RX:=90;
    If RX<-90 then RX:=-90;
    If RY>360 then RY:=RY-360;
    If RY<0 then RY:=Ry+360;
    mx:=x;
    my:=y;
  end;
end;

procedure TForm1.FormMouseDown(Sender: TObject; Button: TMouseButton;
  Shift: TShiftState; X, Y: Integer);
begin
  mx:=x;
  my:=y;
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  octree.Destroy;
  wglMakeCurrent(DC,0);
  wglDeleteContext(RC);
  releaseDC(0,DC);
end;

end.

